from pathlib import Path
from tempfile import TemporaryDirectory

from aseprite_reader import AsepriteFile

from sprak.models.atlas import Atlas
from sprak.models.frame import Frame


class SpritePacker:
    def __init__(self) -> None:
        self._atlas_width = 64
        self._atlas_height = 64
        self._atlas_step = 64
        self._source_folders: list[Path] = []
        self._temp_dir: Path | None = None

    def set_atlas_size(self, width: int, height: int) -> None:
        """Set the starting size of the atlas."""
        self._atlas_width = width
        self._atlas_height = height

    def set_atlas_step(self, step: int) -> None:
        """Set the step size that the atlas increments when it increases its size."""
        self._atlas_step = step

    def add_source_folder(self, folder_path: Path) -> None:
        """Add a directory to the list of source paths that will be processed."""
        if not folder_path.is_dir():
            raise NotADirectoryError(f"{folder_path.as_posix()} must be a directory.")
        self._source_folders.append(folder_path)

    def pack(self, output_folder: Path, atlas_name: str = "atlas") -> None:
        """Pack sprites."""
        self._validate_output_folder(output_folder)

        with TemporaryDirectory() as tmp:
            self._temp_dir = Path(tmp)

            frames = self._collect_frames()

            atlas = Atlas()
            atlas.width = self._atlas_width
            atlas.height = self._atlas_height
            atlas.step_size = self._atlas_step
            atlas.add_frames(frames)
            atlas.write_image(output_folder / f"{atlas_name}.png")
            atlas.write_frame_data(output_folder / f"{atlas_name}.framedata")

            self._temp_dir = None

    @staticmethod
    def _validate_output_folder(output_folder: Path) -> None:
        """Make sure output folder is valid."""
        if not output_folder.is_dir():
            raise NotADirectoryError(f"{output_folder.as_posix()} must be a directory.")

    def _collect_frames(self) -> list[Frame]:
        """Gather a list of frames from the source folders."""
        frames: list[Frame] = []
        for source_folder in self._source_folders:
            for root, dirs, files in source_folder.walk():
                for f in files:
                    file = root / f
                    match file.suffix.lower():
                        case ".png":
                            png_frame = self._png_to_frame(file)
                            frames.append(png_frame)
                        case ".aseprite":
                            aseprite_frames = self._aseprite_to_frames(file)
                            frames += aseprite_frames
        return frames

    def _png_to_frame(self, png_file_path: Path) -> Frame:
        """Create a frame from a PNG file."""
        frame_name = self._rel_path_without_extension(png_file_path)
        sprite_name = self._rel_path_without_extension(png_file_path)
        frame = Frame(frame_name, png_file_path)
        frame.metadata.update({"source_format": "png"})
        frame.metadata.update({"sprite_name": sprite_name})
        return frame

    def _aseprite_to_frames(self, aseprite_file_path: Path) -> list[Frame]:
        """Create frames from an Aseprite file.
        The creation of frames will differ based on whether there are multiple frames or tags in the Aseprite file.
        """
        if not self._temp_dir:
            raise RuntimeError("Temp dir is not set")
        elif not self._temp_dir.exists():
            raise FileNotFoundError(self._temp_dir.as_posix())

        frames = []
        aseprite_file = AsepriteFile(aseprite_file_path)

        sprite_name = self._rel_path_without_extension(aseprite_file_path)
        png_path = self._temp_dir / self._rel_path_to_source_folder(aseprite_file_path).parent
        png_path.mkdir(parents=True, exist_ok=True)

        # Render PNGs
        for frame_number, ase_frame in aseprite_file.iter_frames():
            # Render PNG
            png_file = png_path / f"{aseprite_file_path.stem}.{frame_number:04d}.png"
            aseprite_file.render(ase_frame, png_file)

            # Create a frame for the atlas
            frame_name = self._rel_path_without_extension(aseprite_file.file_path.with_name(png_file.name))
            frame = Frame(frame_name, png_file)

            # Get a list of tags for the frame
            tags = [t.name for t in aseprite_file.frame_tags(frame_number)]

            # Add metadata
            frame.metadata.update({"source_format": "aseprite"})
            frame.metadata.update({"sprite_name": sprite_name})
            frame.metadata.update({"tags": tags})
            frame.metadata.update({"duration": ase_frame.duration})
            frame.metadata.update({"frame_number": frame_number})

            # Add frame to list
            frames.append(frame)

        return frames

    def _rel_path_to_source_folder(self, file: Path) -> Path:
        """Convert a file path to a path relative to its source folder."""
        for source_folder in self._source_folders:
            if file.is_relative_to(source_folder):
                return file.relative_to(source_folder)

        raise RuntimeError(f"File is not relative to any source folders: {file.as_posix()}")

    def _rel_path_without_extension(self, file: Path) -> str:
        """Convert a file path to a relative path string, without an extension.
        The path is relative to the source folder.

        Examples

        File Name: C:/source_folder/Sprite.png
        Frame Name: Sprite

        File Name C:/source_folder/animation/Sprite.0001.png
        Frame Name: animation/Sprite.0001
        """
        # Get the relative path to the file from its source folder
        rel_path = self._rel_path_to_source_folder(file)

        # Strip extension
        no_ext = rel_path.with_suffix("")

        # Convert to forward slashes
        frame_name = no_ext.as_posix()
        return frame_name



from __future__ import annotations
from dataclasses import dataclass


@dataclass
class Rect:
    x: int
    y: int
    w: int
    h: int

    @property
    def top(self) -> int:
        return self.y

    @property
    def bottom(self) -> int:
        return self.y + self.h - 1

    @property
    def left(self) -> int:
        return self.x

    @property
    def right(self) -> int:
        return self.x + self.w - 1

    @property
    def is_empty(self) -> bool:
        """Check if this rectangle has an area of zero."""
        return self.w == 0 or self.h == 0

    def intersects(self, other: Rect) -> bool:
        """Check if this rectangle intersects another rectangle."""
        return (
            other.left <= self.right
            and self.left <= other.right
            and other.top <= self.bottom
            and self.top <= other.bottom
        )

    def as_tuple(self) -> tuple[int, int, int, int]:
        """Return a tuple in the form of (left, top, right, bottom)."""
        return self.left, self.top, self.right, self.bottom



from pathlib import Path
from typing import Any

from PIL import Image


class Frame:
    def __init__(self, name: str, source_image: Path) -> None:
        self._image = None

        # The name of the frame - referencing this is how we'll retrieve it from the atlas
        self.name: str = name

        # Original sprite size
        self.sprite_width: int = 0
        self.sprite_height: int = 0

        # The sprite's top-left pixel offset in the original canvas
        self.offset_x: int = 0
        self.offset_y: int = 0

        # Frame width after trimming transparent edges
        self.frame_width: int = 0
        self.frame_height: int = 0

        # Frame position on the atlas
        self.x: int = 0
        self.y: int = 0

        # Extra metadata that will be included in the framedata file
        self.metadata: dict[str, Any] = {}

        self.load_source_image(source_image)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def image(self) -> Image.Image:
        """The cropped image that will be placed onto the atlas."""
        return self._image

    def load_source_image(self, source_image: Path) -> None:
        """Load source image data."""
        with Image.open(source_image) as image:
            self.sprite_width = image.width
            self.sprite_height = image.height
            self.frame_width = image.width
            self.frame_height = image.height
            self._image = image

            if bbox := image.getbbox():
                left, upper, right, lower = bbox
                self.offset_x = left
                self.offset_y = upper
                self.frame_width = right - left
                self.frame_height = lower - upper
                self._image = image.crop(bbox)


    def to_dict(self) -> dict:
        """Return the frame as a dictionary that can be JSON serialized."""
        return {
            "name": self.name,
            "x": self.x,
            "y": self.y,
            "frame_width": self.frame_width,
            "frame_height": self.frame_height,
            "offset_x": self.offset_x,
            "offset_y": self.offset_y,
            "sprite_width": self.sprite_width,
            "sprite_height": self.sprite_height,
            "metadata": self.metadata,
        }



import json
from pathlib import Path
from typing import Optional

from PIL import Image

from sprak.models.frame import Frame
from sprak.models.rect import Rect


class Atlas:
    def __init__(self) -> None:
        # Tracks whether or not the frames have been packed
        self._frames_packed: bool = False

        # The list of empty rectangular regions that can store frames
        self._regions: list[Rect] = []

        # Stores the length of the progress message so it can be cleared
        self._progress_message_length = 0

        # The size of the atlas
        self.width: int = 64
        self.height: int = 64

        # The amount that the atlas increases in size
        self.step_size: int = 64

        # The list of frames in the atlas
        self.frames: list[Frame] = []

    def add_frames(self, frames: list[Frame]) -> None:
        """Add multiple frames to the atlas."""
        self.frames += frames

    def pack(self) -> None:
        """Pack all frames into the atlas."""
        self.sort_frames_by_height()
        while True:
            if self.place_frames():
                break
            else:
                self.increase_atlas_size()

        self._frames_packed = True

    def sort_frames_by_height(self) -> None:
        """Sort frames by height.
        This seems to leave the least amount of wasted space when we split regions horizontally.
        """
        self.frames.sort(key=lambda f: f.frame_height, reverse=True)

    def place_frames(self) -> bool:
        """Place all frames onto the atlas.
        If the function is successful (that is, it places all frames without running out of space), it will return True.
        Otherwise, it will return False.
        """
        self.reset_regions()
        for frame in self.frames:
            # Skip completely transparent images
            if not frame.image.getbbox():
                continue

            # Find a region that the frame fits in
            if region := self.find_region(frame):
                self.place_frame(frame, region)
            else:
                return False

        return True

    def reset_regions(self) -> None:
        """Reset the list of free regions."""
        self._regions = [Rect(0, 0, self.width, self.height)]

    def find_region(self, frame: Frame) -> Optional[Rect]:
        """Find a free region to pack a frame.
        If no valid position is found, None is returned.
        """
        for region in self._regions:
            w = frame.frame_width
            h = frame.frame_height
            if w <= region.w and h <= region.h:
                return region

        return None

    def place_frame(self, frame: Frame, region: Rect) -> None:
        """Place a frame onto the atlas."""
        frame.x = region.x
        frame.y = region.y

        # Split the leftover space in the region
        split_y = frame.y + frame.frame_height
        split_x = frame.x + frame.frame_width
        self.split_region(region, split_y, split_x)

        # Sort regions from smallest to largest area
        self._regions.sort(key=lambda r: r.w * r.h)

    def split_region(self, region: Rect, y: int, x: int) -> None:
        """Horizontally split a region into sub-regions.
                X
        +------+--------+
        |######|   top  |
        |######|        |
        +------+--------|
        |     bottom    | Y
        |               |
        +---------------+
        """
        # First remove the existing region
        self._regions.remove(region)

        # Top region
        top_w = region.right - x + 1
        top_h = y - region.top
        top = Rect(x, region.y, top_w, top_h)

        # Bottom region
        bottom_w = region.w
        bottom_h = region.bottom - y + 1
        bottom = Rect(region.x, y, bottom_w, bottom_h)

        # Add non-empty regions back to region list
        if not top.is_empty:
            self._regions.append(top)
        if not bottom.is_empty:
            self._regions.append(bottom)

    def increase_atlas_size(self) -> None:
        """Increase the size of the atlas."""
        self.width += self.step_size
        self.height += self.step_size

    def write_image(self, image_file: Path) -> None:
        """Write the atlas out to a file."""
        if not self._frames_packed:
            self.pack()

        image = Image.new("RGBA", size=(self.width, self.height))
        for frame in self.frames:
            # Skip completely transparent images
            if not frame.image.getbbox():
                continue

            image.paste(frame.image, box=(frame.x, frame.y))

        image.save(image_file)

    def write_frame_data(self, data_file: Path) -> None:
        """Write the frame data to a file."""
        # Sort frames alphabetically
        frames = sorted(self.frames, key=lambda f: f.name)

        # Convert frame data to dictionary
        frame_data = {}
        for frame in frames:
            frame_data[frame.name] = frame.to_dict()

        # Write frame data to file
        with data_file.open("w") as fp:
            data_str = json.dumps(frame_data, indent=2)
            fp.write(data_str)
